#include "xprt.h"

/* ------------------------------ InitIRLS ------------------------------- */

void InitIRLS(ExampleVector *w, ExampleVector *e, InputVector *rhs, 
              Matrix *mx, int **indx)
{
   *w    = NewVector(gNumPts);
   *e    = NewVector(gNumPts);
   *rhs  = NewVector(gDimX);
   *mx   = NewMatrix(gDimX, gDimX);
   *indx = (int*)(SafeCalloc(gDimX, sizeof(int)));
}

/* -------------------------------- IRLS --------------------------------- */

void IRLS(ExampleVector y, ExampleVector yWts, WeightVector beta, 
          ModelType model, double phi)
{
   ExampleNumber t;
   CoordinateNumber m, n;
   int loop;
   static Vector e, w, rhs;
   static Matrix mx=NULL;
   static int *indx;
   double rphi, eta, xe, link, dink, error, del;

   if (mx==NULL)
      InitIRLS(&w, &e, &rhs, &mx, &indx);
   
   loop = 0;
   do /* while (error > kIrlsErrCrit && ++loop < kIrlsMaxLoops) */
   {
      switch (model) {
         case gaussian:
            rphi = 1.0/phi;
            for (t=0; t<gNumPts; ++t)
            {
               e[t] = y[t];
               for (n=0; n<gDimX; ++n)
                  e[t] -= beta[n] * gX[t][n];

               w[t] = (yWts) ? yWts[t]*rphi : rphi;
            }
            break;
        case bernoulli:
           for (t=0; t<gNumPts; ++t)
           {
              eta = 0.0;
              for (n=0; n<gDimX; ++n)
                 eta += beta[n] * gX[t][n];

              xe = exp(eta); 
              link = xe/(1.0+xe);
              dink = xe/((1.0+xe)*(1.0+xe));

              e[t] = (y[t] - link)/dink;
              w[t] = (yWts) ? yWts[t] * dink : dink;
           }
           break;
      }
		
      for (m=0; m<gDimX; ++m)
         for (n=0; n<gDimX; ++n)
         {
            mx[m][n] = 0.0;
            for (t=0; t<gNumPts; ++t)
               mx[m][n] += w[t] * gX[t][n] * gX[t][m];
         }

      for (m=0; m<gDimX; ++m)
      {
         rhs[m] = 0.0;
         for (t=0; t<gNumPts; ++t)
            rhs[m] += gX[t][m] * w[t] * e[t];

         for (n=0; n<gDimX; ++n)
            rhs[m] += mx[m][n] * beta[n];
      }			

      LuDecomp(mx, gDimX, indx);
      LuBackSub(mx, gDimX, indx, rhs);

      error = 0.0;
      if (model==gaussian)
         for (m=0; m<gDimX; ++m)
            beta[m] = rhs[m]; 
      else
         for (m=0; m<gDimX; ++m)
         {
            del = rhs[m]-beta[m];
            error += del*del;
            beta[m] = rhs[m];
         } 

   } while (error > kIrlsErrCrit && ++loop < kIrlsMaxLoops);
}

/* ------------------------------ LuDecomp -------------------------------- */

void LuDecomp(Matrix mx, int nDim, int *indx) 
{
   int i, imax, j, k;
   double big, dum, sum, temp;
   Vector vect, row;

   vect = NewVector(nDim);

   for (i=0; i<nDim; ++i)
   {
      big = 0.0;
      for (j=0; j<nDim; ++j)
         if ((temp=fabs(mx[i][j])) > big)
            big = temp;
      if (big==0.0)
      {
         PrintMatrix(mx, gDimX, gDimX);
         Error("Singular matrix in routine LuDecomp");
      }
      vect[i] = 1.0/big;
   }

   for (j=0; j<nDim; ++j)
   {
      for (i=0; i<j; ++i)
      {
         sum = mx[i][j];
         for (k=0; k<i; ++k)
            sum -= mx[i][k] * mx[k][j];
         mx[i][j] = sum;
      }

      big = 0.0;
      for (i=j; i<nDim; ++i)
      {
         sum = mx[i][j];
         for (k=0; k<j; ++k)
            sum -= mx[i][k] * mx[k][j];
         mx[i][j] = sum;
         if ((dum=vect[i]*fabs(sum)) >= big)
         {
            big=dum;
            imax=i;
         }
      }

      if (j != imax)
      {
         row = mx[imax];
         mx[imax] = mx[j];
         mx[j] = row;
         vect[imax] = vect[j];
      }
         
      indx[j] = imax;
      if (mx[j][j]==0.0)
      {
         PrintMatrix(mx, gDimX, gDimX);
         Error("Singular matrix in routine LuDecomp");
      }

      if (j != nDim)
      {
         dum = 1.0/mx[j][j];
         for (i=j+1; i<nDim; ++i)
            mx[i][j] *= dum;
      }
   }

   free(vect);
}
         
/* ------------------------------ LuBackSub ------------------------------- */

void LuBackSub(Matrix mx, int nDim, int *indx, Vector rhs)   
{
   int i, ip, j, ii = -1;
   double sum;

   for (i=0; i<nDim; ++i)
   {
      ip = indx[i];
      sum = rhs[ip];
      rhs[ip] = rhs[i];
      if (ii >= 0)
         for (j=ii; j<=i-1; ++j)
            sum -= mx[i][j]*rhs[j];
      else if (sum)
         ii = i;
      rhs[i] = sum;
   }

   for (i=nDim-1; i>=0; --i)
   {
      sum = rhs[i];
      for (j=i+1; j<nDim; ++j)
         sum -= mx[i][j] * rhs[j]; 
      rhs[i] = sum/mx[i][i];
   }
}

/* ------------------------------------------------------------------------- */
